#include <mpi.h>
#include "neocca.hh"
#include "neoports.hh"
#include "neosupport/MPISetup.hh"

#include <iostream>
//---------------------------------------------------------------
//---------------------------------------------------------------


namespace {

class CommHolder {
private:
	MPI_Comm com;
public:
	CommHolder(MPI_Comm c) : com(c) {}
	~CommHolder(){}
	MPI_Comm getComm() { return com; }
	void destroy() {
		MPI_Comm_free(&com);
	}
};



//---------------------------------------------------------------
//---------------------------------------------------------------

class NeoMPIService_Impl : public virtual neo::cca::ports::MPIService
{

private:
	/** coms for recycle */
	std::vector< CommHolder *> rlist;
	/** coms in use */
	std::vector< CommHolder *> ulist;
	MPI_Comm prototype;

public:

	NeoMPIService_Impl(MPI_Comm comm);
	virtual ~NeoMPIService_Impl();
	virtual MPI_Comm getComm();
	virtual void releaseComm(MPI_Comm m);
	void reclaim();

};



//---------------------------------------------------------------
//---------------------------------------------------------------

class NeoMPIBorrow_Impl : public virtual neo::cca::ports::MPIBorrow
{

private:

	/** next available tag. */
	int maxTag;

	/** last issued key */
	int maxKey;
	
	MPI_Comm common;

public:

  NeoMPIBorrow_Impl(MPI_Comm shareable);

  virtual ~NeoMPIBorrow_Impl();

	/** Get a communicator and a set of available tags on it.
	 * FIXME need to keep a key:tag lookup array to be fully safe.
	 * FIXME need to keep a key comm lookup, if we manage more than one comm...
	 */
  virtual MPI_Comm borrowComm(int tagcount, int* tags, int&key);

	/** Return a communicator and release the tags on it. */
  virtual void returnComm(MPI_Comm c, int tagcount, int*tags, int key);

};



//---------------------------------------------------------------
//---------------------------------------------------------------


NeoMPIBorrow_Impl::NeoMPIBorrow_Impl(MPI_Comm comm) {
	common = comm;
	maxTag = 0;
	maxKey = 0;
}

NeoMPIBorrow_Impl::~NeoMPIBorrow_Impl() {
	maxTag = -1;
	maxKey = -1;
}

MPI_Comm 
NeoMPIBorrow_Impl::borrowComm(int tcount, int *tags, int &key) {
	if (this == 0 || maxTag == -1) { return MPI_COMM_NULL; }
	maxKey++;
	key = maxKey;
	for (int i = 0; i < tcount; i++) {
		tags[i] = i+ maxTag;
	}
	maxTag = tags[tcount-1] + 1;
	return common;
}

void NeoMPIBorrow_Impl::returnComm(MPI_Comm comm, int tcount, int *tags, int key) {
	if (this == 0 || maxTag == -1) { return; }
	// so if we get a use/return, we don't 
	if (key == maxKey-1) {
		maxKey--;
	} else {
		if (key < 0 || key > maxKey) { return; }
	}
	if (tcount < 0 || tcount > maxTag) { return; }
	for (int i = 0; i < tcount; i++) {
		tags[i] = -1;
	}
}

//------------------------------------------------
//------------------------------------------------


NeoMPIService_Impl::NeoMPIService_Impl(MPI_Comm comm) {
	prototype = comm;
}

NeoMPIService_Impl::~NeoMPIService_Impl() {
	size_t i, cap;
	cap = rlist.size();
	for (i = 0; i < cap; i++) {
		CommHolder *c = rlist[i];
		c->destroy();	
		delete c;
		rlist[i] = 0;
	}
	rlist.clear();
	cap = ulist.size();
	for (i = 0; i < cap; i++) {
		// this shouldn't occur if user is correct.
		CommHolder *c2 = ulist[i];
		delete c2;
		ulist[i] = 0;
	}
	ulist.clear();
}

void
NeoMPIService_Impl::reclaim() {
	size_t i, cap;
	cap = rlist.size();
	for (i = 0; i < cap; i++) {
		CommHolder *c = rlist[i];
		c->destroy();	
	}
	cap = ulist.size();
	for (i = 0; i < cap; i++) {
		CommHolder *c2 = ulist[i];
		c2->destroy();	
	}
	ulist.clear();
}

MPI_Comm
NeoMPIService_Impl::getComm() {
	// check for one recyclable
	if (rlist.size() > 0) {
		CommHolder *c = rlist[(rlist.size()-1)];
		rlist.erase(rlist.end()-1);
		ulist.push_back(c);
		MPI_Comm result = c->getComm();
		return result;
	}
	// make a new duplicate. save it and return it.
	MPI_Comm tmp;
	MPI_Comm_dup(prototype, &tmp);
	CommHolder *c2 = new CommHolder(tmp);
        ulist.push_back(c2);	
	return tmp;
}

void 
NeoMPIService_Impl::releaseComm(MPI_Comm m) {
  
	if (m == MPI_COMM_NULL) {
		return;
	}
	// if we don't find it, ignore it quietly.
	size_t i, cap;
	cap = ulist.size();
	for (i = 0; i < cap; i++) {
		int result;
		MPI_Comm tmp = ulist[i]->getComm();
		MPI_Comm_compare(m,tmp,&result);
		if (result == MPI_IDENT) {
			CommHolder *c = ulist[i];
			rlist.push_back(c);
			ulist.erase(ulist.begin()+i);
			return;
		}
	}
}


} // end anon namespace

//---------------------------------------------------------------
//---------------------------------------------------------------

neo::support::MPISetup::MPISetup()
{
	ms = 0;
	naf = 0;
	mpib = 0;
	mpis = 0;
	sr = 0;
	finalized = false;
}

neo::support::MPISetup::~MPISetup()
{
	if (!finalized) { finalize(false); }
}

void 
neo::support::MPISetup::initialize(MPI_Comm borrowComm, MPI_Comm dupComm, neo::cca::AbstractFramework *af)
{
	NeoMPIService_Impl *m1 = new NeoMPIService_Impl(dupComm);
	NeoMPIBorrow_Impl *m2 = new NeoMPIBorrow_Impl(borrowComm);
	mpis = m1;
	mpib = m2;
	naf = af;
	if (!naf)
       	{ 
		std::cerr << "Ugh! no framework. dying..." << std::endl;
		throw neo::cca::Exception("MPISetup::initialize got af==0");
	}

	neo::cca::TypeMap_shared mstm;
	ms = naf->getServices("MPISetup","MPISetup", mstm);
	ms->registerUsesPort("serviceRegistry","neo::cca::ports::ServiceRegistry");
	neo::cca::Port *p = ms->getPort("serviceRegistry");
	sr = dynamic_cast< neo::cca::ports::ServiceRegistry *>(p);
	if (!sr)
       	{ 
		throw neo::cca::Exception("no ServiceRegistry available.");
	}

	// We can add the same service under multiple types so that
	// the users of mpi ports don't have to be perfect within
	// the ambiguities of c++.
	// support 3:
	// 	neo::cca::ports::MPIBorrow/Service
	// 	::neo::cca::ports::MPIBorrow/Service
	// 	neo.cca.ports.MPIBorrow/Service
	if (!sr->addSingletonService("neo::cca::ports::MPIBorrow", mpib))
       	{
		throw neo::cca::Exception("ServiceRegistry rejected neo::cca::ports::MPIBorrow.");
	}
	if (!sr->addSingletonService("::neo::cca::ports::MPIBorrow", mpib))
       	{
		throw neo::cca::Exception("ServiceRegistry rejected ::neo::cca::ports::MPIBorrow." );
	}
	if (!sr->addSingletonService("neo.cca.ports.MPIBorrow", mpib))
       	{
		throw neo::cca::Exception("ServiceRegistry rejected neo.cca.ports.MPIBorrow." );
	}
	if (!sr->addSingletonService("neo::cca::ports::MPIService", mpis))
       	{
		throw neo::cca::Exception("ServiceRegistry rejected neo::cca::ports::MPIService." );
	}
	if (!sr->addSingletonService("::neo::cca::ports::MPIService", mpis))
       	{
		throw neo::cca::Exception("ServiceRegistry rejected ::neo::cca::ports::MPIService." );
	}
	if (!sr->addSingletonService("neo.cca.ports.MPIService", mpis))
       	{
		throw neo::cca::Exception("ServiceRegistry rejected neo.cca.ports.MPIService." );
	}
}

void 
neo::support::MPISetup::finalize(bool reclaim)
{

	finalized = true;
	sr->removeService("neo.cca.ports.MPIService");
	sr->removeService("neo::cca::ports::MPIService");
	sr->removeService("::neo::cca::ports::MPIService");
	sr->removeService("neo.cca.ports.MPIBorrow");
	sr->removeService("neo::cca::ports::MPIBorrow");
	sr->removeService("::neo::cca::ports::MPIBorrow");
	ms->releasePort("serviceRegistry");
	ms->unregisterUsesPort("serviceRegistry");
	sr = 0;
	naf->releaseServices(ms);
	ms = 0;

	if (reclaim) {
		NeoMPIService_Impl * m1 = dynamic_cast<NeoMPIService_Impl *>(mpis);
		m1->reclaim();
	}
	delete mpis;
	delete mpib;

	naf = 0;
}

